from datetime import date, datetime

from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.fields.files import FieldFile, ImageFieldFile
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.template.defaultfilters import yesno
from django.utils.formats import date_format
from django.utils.html import escape
from django.utils.safestring import SafeText, mark_safe
from django.utils.translation import ugettext_lazy as _


__all__ = ['BaseSerializer', 'ViewSetSerializer']


class BaseSerializer:
    empty_values = ('', None)
    type_map = {
        bool: 'handle_bool',
        models.Model: 'handle_model',
        models.QuerySet: 'handle_qs',
        models.Manager: 'handle_manager',
        ImageFieldFile: 'handle_image_file',
        FieldFile: 'handle_file',
        datetime: 'handle_datetime',
        date: 'handle_date',
    }

    def __init__(self, object_list=None, instance=None, **kwargs):
        if object_list is None and instance is None:
            raise ValueError('Must define "object_list" or "instance".')
        self.object_list = object_list
        self.instance = instance
        if self.instance:
            self.model = self.instance.__class__
        else:
            self.model = object_list.model
        self.fields = kwargs.get('fields') or ()
        self.fields_names = kwargs.get('fields_names') or {}
        self.default_value = kwargs.get('default_value') or ''
        self.datetime_format = kwargs.get('datetime_format')
        self.date_format = kwargs.get('date_format')

    def __iter__(self):
        if self.instance:
            return self.serialize_instance()
        else:
            return self.serialize_list()

    def get_field_name(self, field):
        if field in self.fields_names:
            return self.fields_names[field]
        try:
            model_field = self.model._meta.get_field(field)
        except FieldDoesNotExist:
            return field
        if isinstance(model_field, ForeignObjectRel):
            related_model = model_field.related_model
            if model_field.one_to_many:
                return related_model._meta.verbose_name_plural
            return related_model._meta.verbose_name
        return model_field.verbose_name

    def get_field_value(self, obj, field):
        if hasattr(self, f'get_{field}_value'):
            return getattr(self, f'get_{field}_value')(obj)
        if hasattr(obj, f'get_{field}_display'):
            return getattr(obj, f'get_{field}_display')()
        return getattr(obj, field)

    def get_typed_value(self, obj, field):
        value = self.get_field_value(obj, field)
        if isinstance(value, str):
            return value
        for _type, handler in self.type_map.items():
            if isinstance(value, _type):
                value = getattr(self, handler)(obj, field, value)
                break
        if callable(value):
            value = value()
        return value

    def get_value(self, obj, field):
        value = self.get_typed_value(obj, field)
        if value in self.empty_values:
            value = self.default_value
        return value

    def serialize_instance(self):
        return ((self.get_field_name(field), self.get_value(self.instance, field)) for field in self.fields)

    def serialize_list(self):
        for obj in self.object_list:
            yield obj, [self.get_value(obj, field) for field in self.fields]

    def handle_model(self, obj, field, value):
        return str(value)

    def handle_qs(self, obj, field, value):
        return ', '.join(map(str, value))

    def handle_manager(self, obj, field, value):
        return self.handle_qs(obj, field, value.all())

    def handle_bool(self, obj, field, value):
        return yesno(value)

    def handle_file(self, obj, field, value):
        if value:
            return value.url

    def handle_image_file(self, obj, field, value):
        return self.handle_file(obj, field, value)

    def handle_datetime(self, obj, field, value):
        return date_format(value, self.datetime_format)

    def handle_date(self, obj, field, value):
        return date_format(value, self.date_format)


class ViewSetSerializer(BaseSerializer):
    def __init__(self, object_list=None, instance=None, **kwargs):
        super().__init__(object_list, instance, **kwargs)
        self.request = kwargs.get('request')
        self.model_viewset = kwargs.get('model_viewset')
        self.bool_as_icon = kwargs.get('bool_as_icon', False)
        self.order_fields = kwargs.get('order_fields') or ()
        self.is_popup = self.request.GET.get('_popup')
        if not self.fields:
            self.fields = ('__str__',)
            self.fields_names = {'__str__': _('Name')}

    def serialize_list(self):
        for obj in self.object_list:
            row = []
            for field in self.fields:
                value = self.get_value(obj, field)
                if value and not row:
                    value = self.get_obj_link(obj, text=str(obj) if field == '__str__' else value)
                row.append(value)
            yield obj, row

    @property
    def cols(self):
        return (
            (field, self.get_field_name(field), self.is_sortable(field))
            for field in self.fields
        )

    def is_sortable(self, field):
        return field in self.order_fields

    def get_model_viewset(self, obj):
        if obj.__class__ is self.model:
            return self.model_viewset
        if not self.model_viewset.viewset:
            return None
        return self.model_viewset.viewset.get_registered(obj.__class__)

    def get_obj_link(self, obj, text=None):
        text = text or str(obj)
        if not isinstance(text, SafeText):
            text = escape(text)

        model_viewset = self.get_model_viewset(obj)
        if not model_viewset:
            return text

        url = model_viewset.get_obj_url(self.request, obj)
        if not url and not self.is_popup:
            return text

        data = ''
        blank = (obj.__class__ is not self.model)
        if self.request.GET.get('_popup'):
            to_field = self.request.GET.get('_to_field') or 'pk'
            data = 'data-popup-value="{}" data-popup-repr="{}" data-popup-url="{}"'.format(
                getattr(obj, to_field, ''), escape(text), model_viewset.get_obj_url(self.request, obj)
            )

        return mark_safe(
            '<a href="{href}" {target} {data}>{text}</a>'.format(
                href=url,
                target='target="_blank"' if blank else '',
                data=data,
                text=text
            )
        )

    def handle_model(self, obj, field, value):
        return self.get_obj_link(value)

    def handle_qs(self, obj, field, value):
        return mark_safe(', '.join(self.get_obj_link(x) for x in value))

    def handle_bool(self, obj, field, value):
        if self.bool_as_icon:
            return mark_safe(
                '<i class="fas fa-{}-circle text-{}"></i>'.format(
                    value and 'check' or 'times', value and 'success' or 'danger'
                )
            )
        return super().handle_bool(obj, field, value)

    def handle_file(self, obj, field, value):
        if value:
            return mark_safe('<a href="{}" target="_blank">{}</a>'.format(value.url, value))



# flake8: noqa
from .serializers import ViewSetSerializer
from .views import (
    ViewSetCreateView, ViewSetDeleteView, ViewSetDetailView, ViewSetIndexView,
    ViewSetListView, ViewSetUpdateView
)
from .viewsets import ModelViewSet, ViewSet


__version__ = '0.1.5'
default_app_config = 'viewsets.apps.ViewsetsConfig'



import copy
import json

from django import forms
from django.conf import settings
from django.contrib.admin.widgets import (
    SELECT2_TRANSLATIONS, url_params_from_lookup_dict
)
from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe
from django.utils.text import Truncator
from django.utils.translation import get_language


class WidgetWrapper(forms.Widget):
    """
    This class is a wrapper to a given widget.
    """
    template_name = 'viewsets/widgets/wrapper.html'

    def __init__(self, widget):
        super().__init__(widget.attrs)
        self.needs_multipart_form = widget.needs_multipart_form
        self.attrs = widget.attrs
        self.widget = widget

    def __deepcopy__(self, memo):
        obj = copy.copy(self)
        obj.widget = copy.deepcopy(self.widget, memo)
        obj.attrs = self.widget.attrs
        memo[id(self)] = obj
        return obj

    @property
    def choices(self):
        return self.widget.choices

    @choices.setter
    def choices(self, value):
        self.widget.choices = value

    def value_from_datadict(self, data, files, name):
        return self.widget.value_from_datadict(data, files, name)

    def value_omitted_from_data(self, data, files, name):
        return self.widget.value_omitted_from_data(data, files, name)

    def id_for_label(self, id_):
        return self.widget.id_for_label(id_)

    def get_context(self, name, value, attrs):
        context = self.widget.get_context(name, value, attrs)
        context['widget_template'] = self.widget.template_name
        return context

    @property
    def is_hidden(self):
        return self.widget.is_hidden

    @property
    def media(self):
        return self.widget.media


class UrlParametersMixin:
    def limit_parameters(self, rel):
        limit_choices_to = rel.limit_choices_to
        if callable(limit_choices_to):
            limit_choices_to = limit_choices_to()
        return url_params_from_lookup_dict(limit_choices_to)

    def get_parameters(self, rel):
        return self.limit_parameters(rel)

    def url_parameters(self, rel):
        params = self.get_parameters(rel)
        if not params:
            return ''
        return '?' + '&amp;'.join('%s=%s' % (k, v) for k, v in params.items())


class RelatedWidgetWrapper(UrlParametersMixin, WidgetWrapper):
    template_name = 'viewsets/widgets/related_wrapper.html'

    def __init__(self, widget, add_related_url='', can_add_related=False):
        super().__init__(widget)
        self.add_related_url = add_related_url
        self.can_add_related = can_add_related

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        url = self.add_related_url
        rel = getattr(self.widget, 'rel', None)
        if rel:
            url += self.url_parameters(rel)
        context.update({
            'add_related_url': url,
            'can_add_related': self.can_add_related
        })
        return context


class LookupWidget(UrlParametersMixin, forms.TextInput):
    template_name = 'viewsets/widgets/lookup.html'

    def __init__(self, rel, list_url, obj_url_fn=None, attrs=None, using=None):
        self.rel = rel
        self.list_url = list_url
        self.obj_url_fn = obj_url_fn
        self.db = using
        super().__init__(attrs)

    def get_context(self, name, value, attrs):
        attrs['class'] = (attrs.get('class', '') + ' js-related-field').strip()
        context = super().get_context(name, value, attrs)
        context['related_url'] = mark_safe(self.list_url + self.url_parameters(self.rel))
        context['widget']['type'] = 'hidden'
        if context['widget']['value']:
            context['link_label'], context['link_url'] = self.label_and_url_for_value(value)
        return context

    def get_parameters(self, rel):
        from django.contrib.admin.views.main import TO_FIELD_VAR
        params = super().get_parameters(rel)
        params.update({TO_FIELD_VAR: self.rel.get_related_field().name})
        return params

    def label_and_url_for_value(self, value):
        key = self.rel.get_related_field().name
        try:
            obj = self.rel.model._default_manager.using(self.db).get(**{key: value})
        except (ValueError, self.rel.model.DoesNotExist, ValidationError):
            return '', ''

        url = ''
        if self.obj_url_fn:
            url = self.obj_url_fn(obj)

        return Truncator(obj).words(14, truncate='...'), url


class AutocompleteMixin(UrlParametersMixin):
    """
    Select widget mixin that loads options via AJAX.
    """

    def __init__(self, rel, url, attrs=None, choices=(), using=None):
        self.rel = rel
        self.url = url
        self.db = using
        self.choices = choices
        self.attrs = {} if attrs is None else attrs.copy()

    def build_attrs(self, base_attrs, extra_attrs=None):
        """
        Set select2's AJAX attributes.
        """
        attrs = super().build_attrs(base_attrs, extra_attrs=extra_attrs)
        attrs.setdefault('class', '')
        attrs.update({
            'data-ajax--cache': 'true',
            'data-ajax--type': 'GET',
            'data-ajax--url': self.url,
            'data-theme': 'bootstrap4',
            'data-allow-clear': json.dumps(not self.is_required),
            'data-placeholder': '',  # Allows clearing of the input.
            'class': attrs['class'] + (' ' if attrs['class'] else '') + 'js-autocomplete',
            'data-params': json.dumps(self.get_parameters(self.rel)),
        })
        return attrs

    def optgroups(self, name, value, attr=None):
        """Return selected options based on the ModelChoiceIterator."""
        default = (None, [], 0)
        groups = [default]
        has_selected = False
        selected_choices = {
            str(v) for v in value
            if str(v) not in self.choices.field.empty_values
        }
        if not self.is_required and not self.allow_multiple_selected:
            default[1].append(self.create_option(name, '', '', False, 0))
        choices = (
            (obj.pk, self.choices.field.label_from_instance(obj))
            for obj in self.choices.queryset.using(self.db).filter(pk__in=selected_choices)
        )
        for option_value, option_label in choices:
            selected = (str(option_value) in value and (has_selected is False or self.allow_multiple_selected))
            has_selected |= selected
            index = len(default[1])
            subgroup = default[1]
            subgroup.append(self.create_option(name, option_value, option_label, selected_choices, index))
        return groups

    @property
    def media(self):
        extra = '' if settings.DEBUG else '.min'
        i18n_name = SELECT2_TRANSLATIONS.get(get_language())
        i18n_file = ('viewsets/js/vendor/select2/i18n/%s.js' % i18n_name,) if i18n_name else ()
        return forms.Media(
            js=(
                'viewsets/js/vendor/select2/select2.full%s.js' % extra,
            ) + i18n_file + (
                'viewsets/js/autocomplete.js',
            ),
            css={
                'screen': (
                    'viewsets/css/vendor/select2/select2%s.css' % extra,
                    'viewsets/css/vendor/select2/select2-bootstrap4%s.css' % extra,
                ),
            },
        )


class AutocompleteWidget(AutocompleteMixin, forms.Select):
    pass


class AutocompleteManyWidget(AutocompleteWidget, forms.SelectMultiple):
    pass



import operator

import django_filters

from django import forms
from django.db import models
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _


class SearchFilter(django_filters.CharFilter):
    operator_choices = {
        'or': operator.or_,
        'and': operator.and_,
    }

    def __init__(self, *args, **kwargs):
        self.search_fields = kwargs.pop('fields', ())
        self.operator = kwargs.pop('operator', 'or')
        self.term_limit = kwargs.pop('term_limit', None)
        kwargs.setdefault('lookup_expr', 'icontains')
        placeholder = kwargs.pop('placeholder', _('Search'))
        super().__init__(*args, **kwargs)
        self.field.widget.attrs['placeholder'] = placeholder

    def get_operator(self):
        return self.operator_choices.get(self.operator)

    def filter(self, qs, value):
        if not value:
            return qs
        query = Q()
        query_operator = self.get_operator()
        terms = value.split()
        if self.term_limit:
            terms = terms[:self.term_limit]
        for term in terms:
            term_q = Q()
            for field_name in self.search_fields:
                term_q |= Q(**{
                    '{}__{}'.format(field_name, self.lookup_expr): term}
                )
            query = query_operator(query, term_q)
        qs = qs.filter(query)
        if self.distinct:
            qs = qs.distinct()
        return qs


class ViewSetFilterSet(django_filters.FilterSet):
    def __init__(self, data=None, queryset=None, *, request=None, prefix=None,
                 search_kwargs=None, order_by=None):
        self.base_filters['o'] = django_filters.OrderingFilter(widget=forms.HiddenInput)
        sp = search_kwargs.pop('parameter', 'q')
        has_search = bool(search_kwargs and search_kwargs.get('fields'))
        if has_search:
            self.base_filters[sp] = SearchFilter(**search_kwargs)
        super().__init__(data, queryset, request=request, prefix=prefix)

        self.order_fields = {}
        if order_by:
            self.get_order_by(order_by)

        self.has_filters = bool(set(self.filters.keys()) - {'o', sp})
        self.search_field = self.form[sp] if has_search else None
        self.filter_fields = [x for x in self.form if x.name not in ('o', sp)]

    def get_order_by(self, order_by):
        model_fields_names = [
            x.name for x in self._meta.model._meta.get_fields()
        ]
        order_by = [
            x for x in order_by
            if x in model_fields_names or isinstance(x, (list, tuple))
        ]
        self.order_fields = self.filters['o'].normalize_fields(order_by)
        self.filters['o'].extra['choices'] = self.filters['o'].build_choices(
            self.order_fields, {}
        )


def viewset_filterset_factory(model, fields=forms.ALL_FIELDS):
    meta = type(
        'Meta', (object,),
        {
            'model': model,
            'fields': fields,
            'filter_overrides': {
                models.CharField: {
                    'filter_class': django_filters.CharFilter,
                    'extra': lambda f: {
                        'lookup_expr': 'icontains',
                    },
                },
            }
        }
    )
    return type(
        '{}FilterSet'.format(model._meta.object_name),
        (ViewSetFilterSet,),
        {'Meta': meta}
    )



from django.db import models
from django.urls import NoReverseMatch, include, path, reverse, reverse_lazy

from viewsets.decorators import (
    add_filterset_formfield_callback, add_form_formfield_callback,
    add_formset_formfield_callback, check_permission_after,
    check_permission_before
)
from viewsets.views import (
    ViewSetCreateView, ViewSetDeleteView, ViewSetDetailView, ViewSetListView,
    ViewSetUpdateView
)
from viewsets.widgets import (
    AutocompleteManyWidget, AutocompleteWidget, LookupWidget,
    RelatedWidgetWrapper
)


class Default:
    pass


class ViewSet:
    """
    The Viewset class is a container for `ModelViewSet`s.
    """
    def __init__(self, namespace='', name='', view_index=None):
        self.namespace = namespace
        self.prefix = '{}:'.format(self.namespace) if self.namespace else ''
        self.name = name or self.namespace
        self.view_index = view_index
        self._registry = {}

    @property
    def urls(self):
        return self.get_urls(), self.namespace, self.name

    @property
    def registered(self):
        return self._registry.items()

    def get_registered(self, model):
        return self._registry.get(model)

    def register(self, model, model_viewset=None, model_namespace=None, **viewset_kwargs):
        """
        Register the given model with the given model_viewset class.
        """
        model_viewset = model_viewset or ModelViewSet
        self._registry[model] = model_viewset(
            model=model, viewset=self, model_namespace=model_namespace, **viewset_kwargs
        )

    def get_urls(self):
        """
        Get all registered urls.
        """
        urls = []
        if self.view_index:
            urls += [path('', self.view_index.as_view(viewset=self), name='index')]
        for model, viewset in self.registered:
            urls += [path('{}/'.format(viewset.url_prefix), include(viewset.urls))]
        return urls

    def get_index_url(self):
        return reverse_lazy('{}index'.format(self.prefix))

    def reverse(self, model, name, *args, **kwargs):
        """
        Get url for the given model.
        """
        try:
            model_viewset = self._registry[model]
            url = model_viewset.reverse(name, *args, **kwargs)
        except (KeyError, NoReverseMatch):
            url = ''
        return url

    def get_obj_url(self, request, obj):
        try:
            model_viewset = self._registry[obj.__class__]
            url = model_viewset.get_obj_url(request, obj)
        except (KeyError, NoReverseMatch):
            url = ''
        return url


class ViewAttrsMixin:
    fields = Default
    fields_names = Default
    serializer = Default
    queryset = Default
    form_class = Default
    formset_classes = Default
    template_name = Default

    filterset_class = Default
    filterset_fields = Default
    filterset_order_by = Default
    autocomplete_filterset_fields = Default

    paginate_by = Default
    autocomplete_paginate_by = Default

    search_fields = Default
    search_kwargs = Default
    autocomplete_search_fields = Default
    autocomplete_search_kwargs = Default

    bool_as_icon = Default
    default_value = Default
    extra_context = Default


class ModelViewSet(ViewAttrsMixin):
    """
    The ModelViewSet class provides CRUD for model.
    """
    view_list = ViewSetListView
    view_create = ViewSetCreateView
    view_update = ViewSetUpdateView
    view_detail = ViewSetDetailView
    view_delete = ViewSetDeleteView

    base_mixin = None
    available_views = None

    lookups = ()
    autocomplete = ()
    formset_lookups = {}
    formset_autocomplete = {}
    filterset_lookups = ()
    filterset_autocomplete = ()

    def __init__(self, model, viewset=None, model_namespace=None, url_prefix=None, **kwargs):
        self.model = model
        self.viewset = viewset
        self.viewset_prefix = viewset and viewset.prefix or ''
        self.opts = model._meta
        self.model_namespace = model_namespace or '{}_{}'.format(self.opts.app_label, self.opts.model_name)
        self.url_prefix = url_prefix or '{}/{}'.format(self.opts.app_label, self.opts.model_name)
        self.available_views = self.available_views or ['list', 'create', 'update', 'detail', 'delete']

        for k, v in kwargs.items():
            setattr(self, k, v)

    def construct_view(self, view):
        """
        Create mixin from base_mixin and add decorators.
        """
        bases = (view,)
        if self.base_mixin:
            bases = (self.base_mixin,) + bases
        mixined_view = type('{}Mixined'.format(view.__name__), bases, {})
        mixined_view = self.add_decorators(mixined_view)
        return mixined_view

    def add_decorators(self, view):
        if hasattr(view, 'check_permission'):
            view.dispatch = check_permission_before(view.dispatch)
            if hasattr(view, 'get_object'):
                view.get_object = check_permission_after(view.get_object)
        if hasattr(view, 'get_form_class'):
            view.get_form_class = add_form_formfield_callback(view.get_form_class)
        if hasattr(view, 'get_formsets'):
            view.get_formsets = add_formset_formfield_callback(view.get_formsets)
        if hasattr(view, 'get_filterset'):
            view.get_filterset = add_filterset_formfield_callback(view.get_filterset)
        return view

    def get_view(self, view, action='', **view_kwargs):
        view = self.construct_view(view)
        return view.as_view(**self.get_view_kwargs(view, action, **view_kwargs))

    def get_view_kwargs(self, view, action, **view_kwargs):
        kwargs = {
            'viewset': self.viewset,
            'model_viewset': self,
            'opts': self.opts,
            'action': action,
        }
        for attr in (x for x in dir(ViewAttrsMixin) if not x.startswith('_')):
            kwargs.setdefault(attr, getattr(self, attr))
            if attr in ['fields', 'fields_names', 'serializer', 'queryset', 'form_class', 'formset_classes',
                        'template_name']:
                prefixed = '{}_{}'.format(action, attr)
                if hasattr(self, prefixed):
                    kwargs[attr] = getattr(self, prefixed)
        kwargs.update(view_kwargs)
        return {k: v for k, v in kwargs.items() if hasattr(view, k) and v is not Default}

    def reverse(self, view_name, *args, **kwargs):
        try:
            return reverse(
                '{}{}:{}'.format(self.viewset_prefix, self.model_namespace, view_name),
                args=args,
                kwargs=kwargs
            )
        except NoReverseMatch:
            return ''

    @classmethod
    def as_crud(cls, **kwargs):
        """
        Initialize ModelViewset instance and return included urlpatterns.
        """
        return include(cls(**kwargs).urls)

    def get_urls(self):
        urls = []
        patterns = (
            ('list', ''),
            ('create', 'add/'),
            ('detail', '<slug:pk>/'),
            ('update', '<slug:pk>/change/'),
            ('delete', '<slug:pk>/delete/')
        )
        for view_name, pattern in patterns:
            if view_name in self.available_views:
                view = getattr(self, 'view_{}'.format(view_name))
                urls.append(path(pattern, self.get_view(view, view_name, model=self.model), name=view_name))
        return urls

    @property
    def urls(self):
        return self.get_urls(), self.model_namespace

    def has_list_permission(self, request):
        return 'list' in self.available_views and request.user.is_authenticated

    def has_create_permission(self, request):
        return 'create' in self.available_views and request.user.is_authenticated

    def has_update_permission(self, request, obj=None):
        return 'update' in self.available_views and request.user.is_authenticated

    def has_delete_permission(self, request, obj=None):
        return 'delete' in self.available_views and request.user.is_authenticated

    def has_detail_permission(self, request, obj=None):
        return 'detail' in self.available_views and request.user.is_authenticated

    def get_obj_url(self, request, obj):
        if self.has_detail_permission(request, obj):
            return self.reverse('detail', pk=obj.pk)
        if self.has_update_permission(request, obj):
            return self.reverse('update', pk=obj.pk)
        return ''

    def formfield_callback(self, db_field, request, lookups=(), autocomplete=(), can_add_related=True, **kwargs):
        if self.viewset and isinstance(db_field, (models.ForeignKey, models.ManyToManyField)):
            url = self.viewset.reverse(db_field.remote_field.model, 'list')
            if isinstance(db_field, models.ManyToManyField) and db_field.name in autocomplete and url:
                kwargs['widget'] = AutocompleteManyWidget(
                    rel=db_field.remote_field,
                    url=url
                )
            elif isinstance(db_field, models.ForeignKey) and db_field.name in lookups and url:
                kwargs['widget'] = LookupWidget(
                    rel=db_field.remote_field,
                    list_url=url,
                    obj_url_fn=lambda obj: self.viewset.get_obj_url(request, obj),
                    using=kwargs.get('using')
                )
            elif db_field.name in autocomplete and url:
                kwargs['widget'] = AutocompleteWidget(rel=db_field.remote_field, url=url)

            if 'widget' in kwargs and can_add_related:
                related_viewset = self.viewset.get_registered(db_field.remote_field.model)
                if related_viewset and related_viewset.has_create_permission(request):
                    kwargs['widget'] = RelatedWidgetWrapper(
                        widget=kwargs['widget'],
                        add_related_url=self.viewset.reverse(db_field.remote_field.model, 'create'),
                        can_add_related=can_add_related,
                    )
        return db_field.formfield(**kwargs)



from functools import partial, wraps

from django import forms


def formset_validate(func):
    @wraps(func)
    def wrapper(self, form):
        if forms.all_valid(self.formsets.values()):
            return func(self, form)
        return self.form_invalid(form)
    return wrapper


def check_permission_before(func):
    """Check permission before calling function"""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        obj = getattr(self, 'object', None)
        if self.check_permission(self.request, obj):
            return func(self, *args, **kwargs)
        return self.permission_denied(obj)
    return wrapper


def check_permission_after(func):
    """Check permission before after function"""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        obj = getattr(self, 'object', None)
        if not self.check_permission(self.request, obj):
            return self.permission_denied(obj)
        return result
    return wrapper


def add_form_formfield_callback(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        form = func(self, *args, **kwargs)
        return forms.modelform_factory(
            model=getattr(form.Meta, 'model', self.model),
            form=form,
            formfield_callback=partial(
                self.model_viewset.formfield_callback,
                request=self.request,
                lookups=self.model_viewset.lookups,
                autocomplete=self.model_viewset.autocomplete,
            )
        )
    return wrapper


def add_formset_formfield_callback(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        formsets = func(self, *args, **kwargs)
        for formset in formsets.values():
            formset.form = forms.modelform_factory(
                model=formset.form.Meta.model,
                form=formset.form,
                formfield_callback=partial(
                    self.model_viewset.formfield_callback,
                    request=self.request,
                    lookups=self.model_viewset.formset_lookups.get(
                        formset.prefix, ()
                    ),
                    autocomplete=self.model_viewset.formset_autocomplete.get(
                        formset.prefix, ()
                    ),
                )
            )
        return formsets
    return wrapper


def add_filterset_formfield_callback(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        filterset = func(self, *args, **kwargs)
        for name, field in filterset.form.fields.items():
            if name in self.model_viewset.filterset_autocomplete \
                    or name in self.model_viewset.filterset_lookups:
                field.widget = self.model_viewset.formfield_callback(
                    db_field=self.model._meta.get_field(name),
                    request=self.request,
                    autocomplete=self.model_viewset.filterset_autocomplete,
                    lookups=self.model_viewset.filterset_lookups,
                    can_add_related=False,
                ).widget
        return filterset
    return wrapper



from django.template import library
from django.template.defaultfilters import unordered_list
from django.utils.http import urlencode


try:
    from django_jinja import library as jinja_library
    import jinja2
except ImportError:
    jinja_library = None

register = library.Library()


@register.simple_tag()
def viewset_url(model_viewset, view_name, *args, **kwargs):
    return model_viewset.reverse(view_name, *args, **kwargs)


@register.simple_tag()
def object_url(viewset, obj, view_name='detail', *args, **kwargs):
    model_viewset = viewset.get_registered(obj.__class__)
    return viewset_url(model_viewset, view_name, *args, **kwargs)


@register.simple_tag(takes_context=True)
def has_perm(context, viewset, perm, *args, **kwargs):
    perm_fn = getattr(viewset, 'has_{}_permission'.format(perm))
    return perm_fn(context['request'], *args, **kwargs)


@register.inclusion_tag('viewsets/includes/_sort_link.html', takes_context=True)
def sort_link(context, field_name, title, sorting):
    if not sorting:
        return {'sorting': False, 'title': title}
    request = context['request']
    filterset = context['filter']
    param = filterset.order_fields.get(field_name)
    order = request.GET.get('o', '')
    order_by = order.lstrip('-')
    direction = ''
    is_ordered = False
    if param == order_by:
        is_ordered = True
        if order == order_by:
            direction = '-'
    return {
        'name': field_name,
        'title': title,
        'sorting': sorting,
        'direction': direction,
        'is_ordered': is_ordered,
        'param': param
    }


@register.simple_tag(takes_context=True)
def url_replace(context, **kwargs):
    query = context['request'].GET.dict()
    query.update(kwargs)
    return urlencode(query)


@register.filter()
def order_formset_forms(formset):
    # Sort in right order if formset has errors
    if formset.can_order and formset.is_bound:
        return sorted(formset, key=lambda x: x.cleaned_data.get('ORDER', None))
    return formset


@register.inclusion_tag('viewsets/includes/_paginator.html', takes_context=True)
def show_paginator(context):
    page_obj = context['page_obj']
    number = page_obj.number
    num_pages = page_obj.paginator.num_pages
    before = [x for x in range(number - 2, number) if x > 0]
    after = [x for x in range(number + 1, number + 3) if x <= num_pages]
    context.update({
        'range_pages': before + [number] + after
    })
    return context


@register.simple_tag(takes_context=True)
def preserve_params(context, encoded=False):
    request = context['request']
    params = request.GET.dict().copy()
    filterset = context.get('filter')
    if filterset:
        params = {k: v for k, v in params.items() if k not in filterset.filters}
    if encoded:
        return urlencode(params)
    return params


@register.filter()
def is_widget_type_of(field, widget_type):
    return hasattr(field.field.widget, 'input_type') and field.field.widget.input_type == widget_type


@register.filter()
def add_css_class(field, css_class):
    classes = field.field.widget.attrs.get('class', '')
    field.field.widget.attrs['class'] = ' '.join([classes, css_class]).strip()
    return field


@register.filter()
def add_error_css_class(field, css_class):
    if field.errors:
        add_css_class(field, css_class)
    return field


if jinja_library:
    @jinja_library.global_function(name='show_paginator')
    @jinja_library.render_with('viewsets/includes/_paginator.html')
    @jinja2.contextfunction
    def _show_paginator(context):
        return show_paginator(dict(context))

    @jinja_library.global_function(name='sort_link')
    @jinja_library.render_with('viewsets/includes/_sort_link.html')
    @jinja2.contextfunction
    def _sort_link(context, field_name, title, sorting):
        return sort_link(dict(context), field_name, title, sorting)

    jinja_library.global_function(viewset_url)
    jinja_library.global_function(object_url)
    jinja_library.global_function(jinja2.contextfunction(url_replace))
    jinja_library.global_function(jinja2.contextfunction(preserve_params))
    jinja_library.filter(order_formset_forms)
    jinja_library.filter(unordered_list)
    jinja_library.filter(fn=add_css_class, name='add_css_class')
    jinja_library.filter(fn=is_widget_type_of, name='is_widget_type_of')
    jinja_library.filter(fn=add_error_css_class, name='add_error_css_class')



from django.contrib import messages
from django.contrib.auth import get_permission_codename
from django.core.exceptions import PermissionDenied
from django.db.transaction import atomic

from viewsets.decorators import formset_validate
from viewsets.serializers import ViewSetSerializer


class SuccessMessageMixin:
    """
    Add a success message.
    """
    success_message = ''

    def get_success_message(self):
        return self.success_message

    def add_message(self):
        success_message = self.get_success_message()
        if success_message:
            messages.success(self.request, success_message)


class FormsetMeta(type):
    """Add decorator to form_valid for checking formsets"""

    def __init__(cls, name, bases, nmspc):
        super(FormsetMeta, cls).__init__(name, bases, nmspc)
        cls.form_valid = formset_validate(cls.form_valid)


class FormsetViewMixin(metaclass=FormsetMeta):
    formset_classes = ()
    sortable_field_names = {}
    default_sortable_field_name = 'position'
    formsets = None

    def get_formset_classes(self):
        return self.formset_classes

    def get_formsets(self):
        formsets = {}
        for formset_class in self.get_formset_classes():
            formset = formset_class(**self.get_formset_kwargs())
            formset.opts = formset.model._meta
            formsets[formset.prefix] = formset
        return formsets

    def get_form(self):
        self.formsets = self.get_formsets()
        return super().get_form()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        form_and_formsets_media = context['form'].media
        for formset in self.formsets.values():
            form_and_formsets_media += formset.media
        context.update({
            'formsets': self.formsets,
            'form_and_formsets_media': form_and_formsets_media
        })
        return context

    def get_sortable_field(self, prefix):
        return self.sortable_field_names.get(
            prefix, self.default_sortable_field_name
        )

    def formsets_save(self):
        for formset in self.formsets.values():
            formset.instance = self.object
            if formset.can_order:
                sortable_field = self.get_sortable_field(formset.prefix)
                for i, formset_form in enumerate(formset.ordered_forms):
                    if getattr(formset_form.instance, sortable_field) != i:
                        formset_form.has_changed = lambda: True
                    setattr(formset_form.instance, sortable_field, i)
            formset.save()

    @atomic
    def form_valid(self, form):
        result = super().form_valid(form)
        self.formsets_save()
        return result

    def get_formset_kwargs(self):
        kwargs = {}
        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        if getattr(self, 'object', None):
            kwargs['instance'] = self.object
        return kwargs


class PermissionMixin:
    def permission_denied(self, obj=None):
        raise PermissionDenied

    def check_permission(self, request, obj=None):
        raise NotImplementedError

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        if hasattr(self, 'model_viewset'):
            for action in ('create', 'update', 'delete', 'list', 'detail'):
                perm_name = 'has_{}_permission'.format(action)
                perm = getattr(self.model_viewset, perm_name)
                perm_args = [self.request]
                if action not in ('create', 'list'):
                    perm_args.append(getattr(self, 'object', None))
                context[perm_name] = perm(*perm_args)
        return context


class SerializerMixin:
    serializer = ViewSetSerializer
    bool_as_icon = True
    default_value = ''

    def get_serializer_class(self):
        return self.serializer

    def get_serializer_kwargs(self, **kwargs):
        order_fields = ()
        if hasattr(self, 'filterset'):
            order_fields = getattr(self.filterset, 'order_fields', ())
        kwargs.update({
            'request': self.request,
            'model_viewset': self.model_viewset,
            'instance': getattr(self, 'object', None),
            'fields': getattr(self, 'fields', ()),
            'fields_names': getattr(self, 'fields_names', {}),
            'default_value': self.default_value,
            'order_fields': order_fields,
            'bool_as_icon': self.bool_as_icon,
        })
        return kwargs

    def get_serializer(self, **kwargs):
        serializer = self.get_serializer_class()
        if serializer:
            return serializer(**self.get_serializer_kwargs(**kwargs))

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['serializer'] = self.get_serializer(
            object_list=context.get('object_list')
        )
        return context


class DjangoPermissionMixin:
    def has_list_permission(self, request):
        codename_view = get_permission_codename('view', self.opts)
        return request.user.has_perm('%s.%s' % (self.opts.app_label, codename_view))

    def has_create_permission(self, request):
        codename = get_permission_codename('add', self.opts)
        return request.user.has_perm('%s.%s' % (self.opts.app_label, codename))

    def has_update_permission(self, request, obj=None):
        codename = get_permission_codename('change', self.opts)
        return request.user.has_perm('%s.%s' % (self.opts.app_label, codename))

    def has_delete_permission(self, request, obj=None):
        codename = get_permission_codename('delete', self.opts)
        return request.user.has_perm('%s.%s' % (self.opts.app_label, codename))

    def has_detail_permission(self, request, obj=None):
        codename_view = get_permission_codename('view', self.opts)
        codename_change = get_permission_codename('change', self.opts)
        return (
            request.user.has_perm('%s.%s' % (self.opts.app_label, codename_view)) or
            request.user.has_perm('%s.%s' % (self.opts.app_label, codename_change))
        )



from distutils.core import setup

from setuptools import find_packages

from viewsets import __version__


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='django-view-sets',
    version=__version__,
    description='Yet another app for making CRUD.',
    long_description=readme(),
    keywords=['django', 'views', 'viewset', 'crud'],
    author='Grigory Mishchenko',
    author_email='grishkokot@gmail.com',
    url='https://github.com/kindlycat/django-view-sets/',
    packages=find_packages(exclude=('manage', 'tests', 'tests.*')),
    include_package_data=True,
    install_requires=['Django>=2.0', 'django-filter>=2.1.0'],
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'License :: OSI Approved :: BSD License',
    ],
)



from django_filters.views import BaseFilterView

from django.contrib.admin.utils import NestedObjects
from django.contrib.admin.widgets import url_params_from_lookup_dict
from django.db import router
from django.db.transaction import atomic
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.urls import NoReverseMatch
from django.utils.encoding import force_text
from django.utils.html import format_html
from django.utils.text import capfirst
from django.utils.translation import ugettext_lazy as _
from django.views.generic import (
    CreateView, DeleteView, DetailView, ListView, TemplateView, UpdateView
)

from viewsets.decorators import check_permission_before
from viewsets.filters import ViewSetFilterSet, viewset_filterset_factory
from viewsets.mixins import (
    FormsetViewMixin, PermissionMixin, SerializerMixin, SuccessMessageMixin
)


__all__ = [
    'BaseViewSetView', 'ViewSetListView',
    'ViewSetCreateView', 'ViewSetUpdateView', 'ViewSetDeleteView',
    'ViewSetDetailView', 'ViewSetIndexView'
]


class BaseViewSetView(PermissionMixin):
    viewset = None
    model_viewset = None
    model = None
    opts = None
    action = None
    is_popup = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.action in ('create', 'update', 'detail') and not getattr(self, 'form_class', None):
            self.fields = self.fields or [x.name for x in self.opts.get_fields() if x.editable and not x.auto_created]

    def dispatch(self, request, *args, **kwargs):
        self.is_popup = bool(self.request.GET.get('_popup') or self.request.POST.get('_popup'))
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        context.update({
            'opts': self.opts,
            'viewset': self.viewset,
            'model_viewset': self.model_viewset,
            'is_popup': self.is_popup,
            'to_field': self.request.GET.get('_to_field', ''),
        })
        return context

    def get_template_names(self):
        if self.template_name:
            return self.template_name
        if self.viewset and self.action:
            template_names = [
                '{}/{}/{}_{}.html'.format(
                    self.viewset.namespace, self.opts.app_label, self.opts.model_name, self.action
                ),
                '{}/{}_{}.html'.format(self.viewset.namespace, self.opts.app_label, self.action),
                '{}/{}.html'.format(self.viewset.namespace, self.action)
            ]
        else:
            template_names = super().get_template_names()
        return template_names + ['viewsets/{}.html'.format(self.action)]

    def get_success_url(self):
        if self.request.POST.get('_continue'):
            return self.model_viewset.reverse('update', self.object.pk)
        if self.request.POST.get('_addanother'):
            return self.model_viewset.reverse('create')
        return self.model_viewset.reverse('list')


class ViewSetListView(SerializerMixin, BaseViewSetView, BaseFilterView, ListView):
    fields = None
    fields_names = None

    filterset_fields = ()
    filterset_order_by = None
    autocomplete_filterset_fields = ()

    paginate_by = 10
    autocomplete_paginate_by = 10

    search_fields = ()
    search_kwargs = {}
    autocomplete_search_fields = ()
    autocomplete_search_kwargs = {}

    as_json = None

    def check_permission(self, request, obj=None):
        return self.model_viewset.has_list_permission(self.request)

    def dispatch(self, request, *args, **kwargs):
        self.as_json = (request.GET.get('as') == 'json')
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        qs = super().get_queryset()
        if self.is_popup or self.as_json and self.request.GET:
            qs = self.filter_lookup(qs)
        return qs

    def filter_lookup(self, qs):
        lookups = []
        for fk_lookup in self.model._meta.related_fkey_lookups:
            if callable(fk_lookup):
                fk_lookup = fk_lookup()
            lookups.append(url_params_from_lookup_dict(fk_lookup))
        for key, value in self.request.GET.items():
            if {key: value} in lookups:
                qs = qs.complex_filter({key: value})
        return qs

    def get_paginate_by(self, queryset):
        if self.as_json:
            return self.autocomplete_paginate_by
        return super().get_paginate_by(queryset)

    def get_filterset_class(self):
        if self.filterset_class:
            return self.filterset_class
        fields_var = 'autocomplete_filterset_fields' if self.as_json else 'filterset_fields'
        filterset_fields = getattr(self, fields_var, ())
        return viewset_filterset_factory(model=self.model, fields=filterset_fields)

    def get_search_kwargs(self):
        kwargs = (self.autocomplete_search_kwargs if self.as_json else self.search_kwargs).copy()
        kwargs['fields'] = self.autocomplete_search_fields if self.as_json else self.search_fields
        return kwargs

    def get_filterset_kwargs(self, filterset_class):
        kwargs = super().get_filterset_kwargs(filterset_class)
        if issubclass(filterset_class, ViewSetFilterSet):
            kwargs.update({
                'order_by': self.filterset_order_by if self.filterset_order_by is not None else self.fields,
                'search_kwargs': self.get_search_kwargs()
            })
        return kwargs

    def get_json_response(self, context):
        data = {
            "results": [{
                "id": obj.pk,
                "text": str(obj)
            } for obj in context.get('object_list', ())]
        }
        if context.get('page_obj'):
            data.update({
                'pagination': {
                    'more': context['page_obj'].has_next()
                }
            })
        return JsonResponse(data)

    def render_to_response(self, context, **response_kwargs):
        if self.as_json:
            return self.get_json_response(context)
        return super().render_to_response(context, **response_kwargs)


class ViewSetDetailView(SerializerMixin, BaseViewSetView, DetailView):
    fields = None
    fields_names = None

    def check_permission(self, request, obj=None):
        return self.model_viewset.has_detail_permission(request, obj)


class ViewSetCreateView(SuccessMessageMixin, FormsetViewMixin, BaseViewSetView, CreateView):
    success_message = _('Object was created.')

    def check_permission(self, request, obj=None):
        return self.model_viewset.has_create_permission(request)

    def form_valid(self, form):
        result = super().form_valid(form)
        if self.is_popup:
            return render(
                self.request,
                'viewsets/popup_response.html',
                context={
                    'object': self.object,
                    'obj_url': self.model_viewset.get_obj_url(self.request, self.object)
                },
            )
        else:
            self.add_message()
        return result


class ViewSetUpdateView(SuccessMessageMixin, FormsetViewMixin, BaseViewSetView, UpdateView):
    success_message = _('Object was updated.')

    def check_permission(self, request, obj=None):
        return self.model_viewset.has_update_permission(request, obj)

    def form_valid(self, form):
        result = super().form_valid(form)
        if not self.is_popup:
            self.add_message()
        return result


class ViewSetDeleteView(BaseViewSetView, SuccessMessageMixin, DeleteView):
    success_message = _('Object was deleted.')

    def check_permission(self, request, obj=None):
        return self.model_viewset.has_delete_permission(request, obj)

    def get_to_delete(self, objs):
        collector = NestedObjects(using=router.db_for_write(self.model))
        collector.collect(objs)
        perms_needed = set()

        def format_callback(obj):
            opts = obj._meta
            no_edit_link = '{}: {}'.format(capfirst(opts.verbose_name), force_text(obj))
            if not self.viewset:
                return no_edit_link

            viewset = self.viewset.get_registered(obj.__class__)

            if viewset:
                try:
                    url = viewset.reverse('update', pk=obj.pk)
                except NoReverseMatch:
                    return no_edit_link

                if not viewset.has_delete_permission(self.request, obj):
                    perms_needed.add(opts.verbose_name)
                return format_html('{}: <a href="{}">{}</a>', capfirst(opts.verbose_name), url, obj)
            else:
                return no_edit_link

        to_delete = collector.nested(format_callback)
        protected = [format_callback(obj) for obj in collector.protected]
        model_count = {model._meta.verbose_name_plural: len(objs) for model, objs in collector.model_objs.items()}
        return to_delete, model_count, perms_needed, protected

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        to_delete, model_count, perms_needed, protected = self.get_to_delete([self.object])
        context.update({
            'to_delete': to_delete,
            'model_count': model_count,
            'perms_needed': perms_needed,
            'protected': protected,
            'object_name': force_text(self.opts.verbose_name)
        })
        return context

    @atomic
    def delete(self, request, *args, **kwargs):
        self.object = self.get_object()
        to_delete, model_count, perms_needed, protected = self.get_to_delete([self.object])
        if perms_needed or protected:
            return redirect('.')
        self.object.delete()
        self.add_message()
        return redirect(self.get_success_url())


class ViewSetIndexView(PermissionMixin, TemplateView):
    template_name = 'viewsets/index.html'
    viewset = None

    @check_permission_before
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def check_permission(self, request, obj=None):
        return request.user.is_authenticated
